#include <assert.h>
#include <stdlib.h>  // for getenv()

#include "config.h"
#include "tree.h"
#include <stdio.h>
#include "cp-tree.h"
#include "tivoDeclIter.h"
#include "tivoRefCount.h"
#include "tivoTable.h"
#include "tivoUtil.h"

#include <string.h>

// HACK: ugly reach into toplevel.c so we can temporarily
//       disable warnings_are_errors if we want to.
extern int warnings_are_errors;

extern int warn_tivo;

/****************************************************************
 *
 * Helpers
 *
 ****************************************************************/

// returns: true iff we should do tivoRefCount checking
//          to avoid the checkng,
//		setenv TIVO_GCC_REF_COUNT_DISABLE someValue
static tivoBool tivoShouldCheck()
{
    return warn_tivo;
}

// returns: true iff we should do tivoRefCount warnings should
//          be errors.  to avoid treating tivoRefCount warnings
//          as errors (even with -Werror), 
//             setenv TIVO_GCC_REF_COUNT_NOT_ERROR
static tivoBool tivoShouldPreventError()
{
    static tivoBool isSet        = tivoFalse;
    static tivoBool preventError = tivoFalse;

    if ( !isSet ) {
	if ( getenv( "TIVO_GCC_REF_COUNT_NOT_ERROR" ) != NULL ) {
	    preventError = tivoTrue;
	}
	isSet = tivoFalse;
    }

    return preventError;
}

// returns: "true" or "false"
static const char* tivoBoolString( tivoBool b )
{
    return b ? "true" : "false";
}

// returns: tivoTrue iff varDecl represents "this"
// requires: varDecl is a VAR_DECL
static tivoBool tivoIsVarDeclThis( tree varDecl )
{
    static tree idOfThis = NULL;

    if ( idOfThis == NULL ) {
	idOfThis = get_identifier( "this" );
    }

    assert( varDecl != NULL && TREE_CODE( varDecl ) == VAR_DECL );
    if ( idOfThis == DECL_NAME(varDecl) ) {
	return tivoTrue;
    }
    else {
	return tivoFalse;
    }
}

tivoBool tivoIsProbablyInstantiationOfRef( const char* name )
{
    const char *s = name;

    // Check if name is a template.
    while (s[0] != '<' && s[0] != 0) s++;
    if (s[0] == 0) return tivoFalse;
    while (s[0] != '>' && s[0] != 0) s++;
    if (s[0] != '>' || s[1] != 0) return tivoFalse;

    // We accept any template name of the form:
    //   ref<>
    //   ref[A-Z].*<>
    //   .*Ref<>
    //   .*Ref[A-Z].*<>

    // Check if name looks like a ref.
    if (strncmp(name, "ref", 3) == 0 &&
            (name[3] == '<' || (name[3] >= 'A' && name[3] <= 'Z'))) {
        return tivoTrue;
    }

    // Does name contain "Ref" followed by a '<' or a capital letter?
    //
    // Note:  We know this string contains a '<' since we already checked for
    // it.  So, we use that sentinel to terminate our loop.
    s = name;
    while (s[0] != '<') {
        while (s[0] != '<' && s[0] != 'R') {
            s++;
        }
        if (s[0] == 'R') {
            if (s[1] == 'e' && s[2] == 'f' && 
                    (s[3] == '<' || (s[3] >= 'A' && s[3] <= 'Z'))) {
                return tivoTrue;
            }
            s++;
        }
    }
    return tivoFalse;
}

/* a nameTypePair is used to pair class names with
 * pointers to the types.
 */
struct _tivoNameTypePair {
    const char* name;
    tree        type;
};

typedef struct _tivoNameTypePair tivoNameTypePair;

/* returns: an array of nameTypePairs which store pointers to
 *          RECORD_TYPE trees for each ref counting base class.  
 *          array is terminated by a nameTypePair with a NULL name.
 *
 * WARNING: if we have not yet found a definition for a given baseClass,
 *          it's type member will still be NULL. 
 *
 * IMPL NOTE: I'm assuming that once we've seen a class, it's record doesn't change!
 */
static tivoNameTypePair* tivoGetRefCountingBaseClasses()
{
    static tivoNameTypePair refCountingBaseClasses[] = {
	{ "TmkCore", NULL },
	{ NULL, NULL }
    };
    static tivoBool foundAll = tivoFalse;

    /* if we haven't found all the types yet, look up missing ones! */
    if ( !foundAll ) {
	tivoBool          stillMissingOne = tivoFalse;
	tivoNameTypePair* scan            = NULL;

	for( scan = refCountingBaseClasses; scan->name; scan++ ) {
	    if ( scan->type == NULL ) {
		/* this type's missing...look for it! */
		tree candidate = lookup_name( get_identifier( (char*) scan->name ),
					  1 /* we're not lex -- see decl.c! */ );

		if ( candidate != NULL ) {
		    if ( TREE_CODE(candidate) == TYPE_DECL ) {
			/* candidate is usually a type declaration, but 
			 * we want to store the actual type node!
			 */
			candidate = TREE_TYPE(candidate);
		    }
		    
		    if ( IS_AGGR_TYPE_CODE( TREE_CODE(candidate) ) ) {
			/* only remember it if it's a class or union! */
			scan->type = candidate;
		    }
		}
	    }
	    if ( scan->type == NULL ) {
		stillMissingOne = tivoTrue;
	    }
	}

	/* if we're not missing any, we've found them all and can skip
	 * the lookups next time!
	 */
	foundAll = !stillMissingOne;
    }

    /* return what we've got -- caller knows we might have NULLs! */
    return refCountingBaseClasses;
}

static tivoBool tivoHasRefCountingBaseClass( tree type )
{
    tivoNameTypePair* scan = NULL;

    if ( !tivoIsClassType( type ) ) {
	assert( tivoFalse && "see requirements!" );
	return tivoFalse;
    }

    for( scan = tivoGetRefCountingBaseClasses(); scan->name; scan++ ) {
	if ( scan->type == NULL ) {
	    /* we haven't seen a non-forward declaration for this puppy yet! */
	    continue;
	}

	if ( DERIVED_FROM_P( scan->type, type ) ) {
	    return tivoTrue;
	}
    }

    return tivoFalse;
}

/****************************************************************
 *
 * Exposed functions
 *
 ****************************************************************/

static tivoBool tivoIsRefCountable( tree type )
{
    char typeCodeClass = TREE_CODE_CLASS( TREE_CODE(type) );

    /* make sure we're working on a type! */
    if ( typeCodeClass != 't' ) {
	assert( tivoFalse && "read requirements!" );
	return tivoFalse;
    }
    
    if ( !tivoIsClassType(type) ) {
	/* if it's not a class, how could it be refCounted? */
	return tivoFalse;
    }

    if ( !tivoHasRefCountingBaseClass(type) ) {
	// isn't subclass of refCounted base class
	return tivoFalse;
    }

    return tivoTrue;
}

/**
 * if node describes a ref type, check that the argument
 * to the template is a subclass of TmkCore.
 */
static void tivoCheckRefRefsReffable( tree node )
{
    tree type = NULL;
    tree templateArgs = NULL;
    tree reffedType = NULL;

    if ( TREE_CODE( node ) != TYPE_DECL ||
         TREE_CODE(TREE_TYPE(node)) != RECORD_TYPE )
    {
        // not a record, nothing to see here.
        return;
    }

    type = TREE_TYPE( node );
    if ( !DECL_USE_TEMPLATE( type ) )
        return;  // not a template.  get out!

    templateArgs = CLASSTYPE_TI_ARGS(type);

    if ( TREE_CODE(templateArgs) != TREE_VEC ||
         TREE_VEC_LENGTH(templateArgs) != 1 )
    {
        // hmmm....must not be the template we're looking for!
        //fprintf( stderr, "looks like a ref<> template instantiation,"
                 //"but doesn't taste like one!\n" );
        return;
    }

    reffedType = TREE_VEC_ELT( templateArgs, 0 );

    if ( 0 == TYPE_SIZE( reffedType ) )
        return;  // can't say whether it's reffable or not -- it's incomplete!

    if ( ! tivoIsRefCountable( reffedType ) )
    {
        int oldWarnAreErrors = warnings_are_errors;
        if ( tivoShouldPreventError() ) { warnings_are_errors = 0; }
        warning_with_file_and_line (DECL_SOURCE_FILE(node),
                                    DECL_SOURCE_LINE(node),
                                    "TiVo-Ref: type being ref<>'d must be a subtype of TmkCore! {%s}",
                                    tivoGetId(reffedType) );
        warnings_are_errors = oldWarnAreErrors;
    }
}

static int tivoCheck( tree node, int fStart )
{
    int nodeCode = ERROR_MARK;
    int retValue = 1;  /* we almost always recurse */

    if ( node == NULL ) {
	// not sure when this happens...but it does!
	return retValue;
    }

    // get the node's code to dispatch below!
    nodeCode = TREE_CODE( node );

    /*
    fprintf( stderr, "---> tivoCheck: %s(%s)\n",
	     tree_code_name[nodeCode],
	     id );
    */

    if ( nodeCode == VAR_DECL ) {
	tree varType       = TREE_TYPE( node );
	tree varNuggetType = tivoGetNuggetType( varType );

	// it turns out that when you inline a method,
	// "this" appears as a local variable.  since you
	// can't ref<> or nonRef<> 'this', we don't want
	// to check it.
	tivoBool isThis    = tivoIsVarDeclThis( node );

	if ( !isThis && tivoIsRefCountable( varNuggetType ) ) {
	    int oldWarnAreErrors = warnings_are_errors;
	    if ( tivoShouldPreventError() ) { warnings_are_errors = 0; }
	    warning_with_file_and_line (DECL_SOURCE_FILE(node),
					DECL_SOURCE_LINE(node),
					"TiVo-Ref: '%s' should be ref<>'d {%s}",
					tivoGetId(node),
					decl_as_string(varNuggetType,1)
					);
	    warnings_are_errors = oldWarnAreErrors;
	}
    }
    else if ( nodeCode == FIELD_DECL ) {
	tree varType       = TREE_TYPE( node );
	tree varNuggetType = tivoGetNuggetType( varType );

	if ( tivoIsRefCountable( varNuggetType ) ) {
	    int oldWarnAreErrors = warnings_are_errors;
	    if ( tivoShouldPreventError() ) { warnings_are_errors = 0; }
	    warning_with_file_and_line( DECL_SOURCE_FILE(node),
					DECL_SOURCE_LINE(node),
					"TiVo-Ref: '%s' should be ref<>'d {%s}",
					tivoGetId(node),
					decl_as_string(varNuggetType,1)
		);
	    warnings_are_errors = oldWarnAreErrors;
	}
    }
    else if ( nodeCode == FUNCTION_DECL ) {
	tree fnType  = TREE_TYPE( node );    // a FUNCTION_TYPE
	tree retType = TREE_TYPE( fnType );  // the return type
	tree retNuggetType = tivoGetNuggetType( retType );

	if ( retNuggetType != NULL ) {
	    // note: constructors appear to return the type, 
	    //       but we don't want to complain about their
	    //       return type!  
	    tivoBool isCtor = ( TYPE_LANG_SPECIFIC(node) != NULL &&
				DECL_CONSTRUCTOR_P(node) );
	    tivoBool isAssignment = (DECL_NAME(node) == get_identifier("__as"));
	    tivoBool considerForRule = (!isCtor && 
					!isAssignment);

	    /*
	    if ( isCtor ) {
		fprintf( stderr, "%s is a constructor!\n", tivoGetId(node) );
	    }
	    if ( isAssignment ) {
		fprintf( stderr, "%s is __as!\n", tivoGetId(node) );
	    }
	    */

	    if ( considerForRule && tivoIsRefCountable( retNuggetType ) ) {
		int oldWarnAreErrors = warnings_are_errors;
		if ( tivoShouldPreventError() ) { warnings_are_errors = 0; }
		warning_with_file_and_line( DECL_SOURCE_FILE(node),
					    DECL_SOURCE_LINE(node),
					    "TiVo-Ref: the return type of '%s' should be refCounted! {%s}", 
					    tivoGetId(node),
					    decl_as_string(retNuggetType,1)
		    );
		warnings_are_errors = oldWarnAreErrors;
	    }
	}
    }
    else if ( nodeCode == TYPE_DECL ) {
	if ( TREE_CODE(TREE_TYPE(node)) == RECORD_TYPE &&
	     tivoIsProbablyInstantiationOfRef( tivoGetId(node) ) ) {
	    // skip instantiations of the ref<> or nonRef<> template!
            if ( fStart )
                tivoCheckRefRefsReffable( node );
            retValue = 0;
	}
    }
    else if ( nodeCode == RECORD_TYPE ) {
    }
    else if ( nodeCode == TREE_VEC ) {
    }
    else {
    }

    return retValue;
}

/**********************************************************
 *
 * High-level checking routines!
 *
 **********************************************************/

extern int expanding_inline_function;

// effects: checks the rules on everything in the current scope
void tivoCheckDecl( tree decl )
{
    if ( !tivoShouldCheck() ) {
	return;
    }

    if ( expanding_inline_function > 0 ) {
	return;
    }

    tivoIterateOver( &tivoCheck, decl );
}

// effects: checks the rules on everything in the current scope
void tivoCheckRefCountingRules()
{
    if ( !tivoShouldCheck() )
        return;

    tivoIterateOverGlobal( &tivoCheck );
}


void tivoCheckForAssignmentToRefCountedParameterType( tree lhs )
{
    tree varType       = NULL;
    tree varNuggetType = NULL;

    if ( TREE_CODE( lhs ) != PARM_DECL )
    {
        // we only check parameters here!
        return;
    }

    varType       = TREE_TYPE( lhs );
    varNuggetType = tivoGetNuggetType( varType );
    
    if ( tivoIsRefCountable( varNuggetType ) )
    {
        int oldWarnAreErrors = warnings_are_errors;
        if ( tivoShouldPreventError() ) { warnings_are_errors = 0; }
        cp_warning( "TiVo-Ref: don't assign to parameter '%s' it's a pointer to a refCountable type! {%s}",
                    tivoGetId( lhs ),
                    decl_as_string(varNuggetType,1) );
        //debug_tree( lhs );
        warnings_are_errors = oldWarnAreErrors;
    }
}

void tivoCheckCall( tree paramlist )
{
    tree p;
            
    if ( !tivoShouldCheck() ) {
	return;
    }

    for ( p = paramlist;  p != NULL;  p = TREE_CHAIN(p) )
    {
        tree arg = TREE_VALUE( p );
        tree argForType = arg;
        
        /* For some reason, function arguments are all wrapped in nop_expr */
        if ( TREE_CODE(arg) == NOP_EXPR )
        {
            argForType = TREE_OPERAND( arg, 0 );
        }
        
        if ( TREE_CODE(argForType) == CALL_EXPR )
        {
            tree argType = TREE_TYPE(arg);
            if ( argType != 0 )
            {
                tree argNuggetType = tivoGetNuggetType( argType );

                // printf( "\n\nparam:\n" );
                // debug_tree( p );

                if ( tivoIsRefCountable( argNuggetType ) )
                {
                    char* str = strdup( type_as_string( argType, 0 ) );
                    cp_warning( 
                        "Passing TmkCore* return value as argument: %s",
                        str );
                    free( str );
                }
            }
        }
    }
}


